@@ -27,6 +27,10 @@
 #include <sys/resource.h>
 #include <unistd.h>
 
+#ifdef	__FreeBSD__
+#  include <sys/ioctl_compat.h>
+#endif	/* __FreeBSD__ */
+
 #if HAVE_SYS_WAIT_H
 # include <sys/wait.h>
 #endif
@@ -66,9 +70,18 @@ gtk_tty_os_get_hintp (void)
     g_warning ("tcgetattr(stdin,) failed: %s", g_strerror (errno));
     
     osdat->tty_termios.c_iflag = BRKINT | ICRNL | IMAXBEL;
-    osdat->tty_termios.c_oflag = CREAD | OPOST | ONLCR | NL0  | CR0 | TAB0 | BS0 | VT0 | FF0;
+
+    osdat->tty_termios.c_oflag = CREAD | OPOST | ONLCR | NL0  | CR0 | TAB0 | BS0 | FF0;
+#ifdef	VT0
+    /* FreeBSD doesn't have it?
+     */
+    osdat->tty_termios.c_oflag |= VT0;
+#endif	/* VT0 */
+
     osdat->tty_termios.c_cflag = CS8;
+
     osdat->tty_termios.c_lflag = ISIG | ICANON | IEXTEN | ECHO | ECHOE | ECHOK | ECHOCTL | ECHOKE;
+
     osdat->tty_termios.c_cc[VERASE] = '\b';
     osdat->tty_termios.c_cc[VKILL] = '\025';
   }
@@ -109,15 +122,30 @@ gtk_tty_os_open_pty (GtkTty *tty)
     {
       *s2 = *suf_2;
       
-      tty->pty_fd = open (pty_name, O_RDWR);
+      tty->pty_fd = open (pty_name, O_RDWR | O_NONBLOCK);
+      if (tty->pty_fd >= 0)
+      {
+	gint tty_fd;
+
+	*pt = 't';
+
+	tty_fd = open (pty_name, O_RDWR);
+	if (tty_fd < 0)
+	{
+	  close (tty->pty_fd);
+	  tty->pty_fd = -1;
+
+	  *pt = 'p';
+	}
+	else
+	  close (tty_fd);
+      }
     }
   }
   
   if (tty->pty_fd < 0)
     return;
   
-  *pt = 't';
-  
   tty->tty_name = g_strdup (pty_name);
 }
 
@@ -296,16 +324,24 @@ static	void
 gtk_tty_os_wait (GtkTty		*tty)
 {
   gint		err;
-  struct rusage res_usage;
   int		status;
   gboolean	restart = FALSE;
+#ifdef	HAVE_WAIT4
+  struct rusage res_usage;
+#endif	/* HAVE_WAIT4 */
   
   g_assert (tty->pid > 0); /* paranoid */
-  
+
+
 wait_restart:
   do
   {
+#ifdef	HAVE_WAIT4
+    memset (&res_usage, 0, sizeof (res_usage));
     err = wait4 (tty->pid, &status, WNOHANG, &res_usage);
+#else	/* not HAVE_WAIT4 */
+    err = waitpid (tty->pid, &status, WNOHANG);
+#endif	/* not HAVE_WAIT4 */
   }
   while (err == -1 && errno == EINTR);
   
@@ -346,10 +382,12 @@ wait_restart:
     }
     else
     {
+#ifdef	HAVE_WAIT4
       tty->sys_usec = res_usage.ru_stime.tv_usec;
       tty->sys_sec = res_usage.ru_stime.tv_sec;
       tty->user_usec = res_usage.ru_utime.tv_usec;
       tty->user_sec = res_usage.ru_utime.tv_sec;
+#endif	/* HAVE_WAIT4 */
     }
   }
 }
